/*
    w_pvrscan.c
    
    This is w_pvrscan, a small utility for channel scanning an ivtv
    supported card.
    Latest versions always at: http://free.pages.at/wirbel4vdr
    
    
    
    
    Copyright (C) 2005 "wirbel" (handygewinnspiel AT gmx DOT de) 

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <unistd.h>
#include <features.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <inttypes.h>
#include <getopt.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <ctype.h>
#include <errno.h>
#include <sys/ioctl.h>
#include <sys/time.h>
#include <math.h>
#include <linux/types.h>
#include <linux/videodev2.h>

#include "w_pvrscan.h"
#include "version.h"				/* generated by 'make version' */

#define __stringify_1(x)	#x
#define __stringify(x)		__stringify_1(x)
#define MAX_CARDS		16		/* max number of /dev/videoN */
#define MAX_TRY			5		/* number of try waiting for tuning lock */
#define RETRY_TIME		125000		/* time between tests for signal strength, sec */
#define TV_TUNING_TIME		1000000		/* time for switching to new channel, sec */
#define RADIO_TUNING_TIME	250000
#define MIN_SIGNAL_STRENGTH	16384		/* 25% of maximum signal strength */
uint debug 			= 0;		/* no debug */
int  verbosity 			= 2;


/* Options */
enum Option {
	OptAdapter	= 'a',
	OptDebug	= 'd',
        OptHelp		= 'h',
	OptOutfile	= 'o',
	OptPlugin	= 'p',
	OptRadio	= 'r',
	OptRadioDev	= 'R',
	OptVdrver	= 'v',
	OptVerbose	= 'V',
	OptQuiet	= 'Q',
};




static struct option long_options[] = {
	/* long command line options */
	{"adapter"	, required_argument	, 0, OptAdapter},
	{"debug"	, no_argument		, 0, OptDebug},
	{"help"		, no_argument		, 0, OptHelp},
	{"output-file"	, required_argument	, 0, OptOutfile},
	{"plugin"	, required_argument	, 0, OptPlugin},
	{"radio"	, no_argument		, 0, OptRadio},
	{"radio-adapter", required_argument	, 0, OptRadioDev},
	{"vdr-version"	, required_argument	, 0, OptVdrver},
	{"verbose"	, no_argument		, 0, OptVerbose},
	{"quiet"	, no_argument		, 0, OptQuiet},
	{0, 0, 0, 0}
};


void usage(void)
{
	info("Usage:\n");
	info("  -p, --plugin        which vdr plugin do you use?\n");
	info("                      ptv:      ptv plugin [default]\n");
	info("                      pvrinput: pvrinput plugin\n");
	info("                      analogtv: analogtv plugin\n");
        info("\n");
	info("  -v, --vdr-version   syntax for vdr version\n");
	info("                      4:        VDR-1.4.x [default]\n");
        info("                      2:        VDR-1.2.x\n");
        info("\n");
	info("  -o, --output-file   output file [default = stdout]\n");
        info("\n");
	info("  -a, --adapter       /dev/videoN [default = auto detect]\n");
        info("\n");
	info("  -r, --radio         search radio channels [default = auto detect]\n");
	info("                      (analogtv doesn't support radio)\n");
        info("\n");
	info("  -R, --radio-adapter /dev/radioN [default = auto detect]\n");
        info("\n");	
	info("  -h, --help          display this help message\n");
	info("  -V, --verbose       verbose(repeat for more)\n");
	info("  -Q, --quiet         quiet(repeat for less)\n");
	exit(0);
}

/* function doioctl, 20051218, wirbel
  encapsulation of ioctl with failure message
  returns non-zero in case of error
*/


static int doioctl(int fd, int request, void *parm, const char *name)
{
  int retVal;
  retVal = ioctl(fd, request, parm);
  if (retVal < 0) {
    error("%s failed: %s\n", name, strerror(errno));
    exit(1);
    }
  return retVal;
}

/* function getSignalstatus, 20051218, wirbel
  returns -1 if video4linux device not opened
           0 if bad  tuner input signal strength
           1 if good tuner signal strength
*/

int getSignalstatus(int fd) {
  struct v4l2_tuner tuner;
  tuner.index=0;
  tuner.reserved[0]=0; tuner.reserved[1]=0; tuner.reserved[2]=0; tuner.reserved[3]=0;
  if (ioctl(fd, VIDIOC_G_TUNER, &tuner)) {
    error("getSignalstatus: VIDIOC_G_TUNER failed\n");
    return -1;
    }
  else {
    if ((debug) && (tuner.signal < MIN_SIGNAL_STRENGTH))
      info(" %d", tuner.signal);
    return (tuner.signal >= MIN_SIGNAL_STRENGTH)?1:0; 
    }
  }

/* function formatFreq, 20060220, wirbel
  returns argument freq rounded in steps of 50
*/

uint formatFreq(uint freq) {
  uint res, rest;
  rest = freq % 50;
  res = (uint) freq / 50;
  if (rest<=25)
	return 50*res;
  else
	return 50*(res+1);

  }


int main(int argc, char **argv)
{
	FILE *conf=stdout;			/* channels.conf output file */
	struct v4l2_ext_controls ext_ctrls;	/* new v4l2 syntax ext ctrls */
	struct v4l2_ext_control ext_ctrl;	/* new v4l2 syntax ext ctrl */
	struct v4l2_capability vcap;    	/* get driver info */
	v4l2_std_id std;                	/* set video standard PAL */
	struct v4l2_format vfmt;		/* set_format/get_format */
	struct v4l2_input vin;			/* list_inputs */
	struct v4l2_output vout;		/* list_outputs */
	struct v4l2_frequency vf;		/* setting frequency */
	struct v4l2_tuner vtuner;		/* radio tuner settings */
	int i,count=0;
	int dev = -1;
	int fd = -1;
	int radio_dev = -1;
	uint tmr;
	uint strength;
        int ofile=0;
	int input;			/* set_input/get_input */
	int output=0;			/* set_output/get_output */
	int option_index = 0;
	int idx = 0;
	int stereo = 0;
	uint vdrver = 3;
	uint radio = 0;
	char ch, device[64],sendername[64];
	char short_options[26 * 2 + 1];
	char *outfile = strdup("");
	char *plugin = strdup("ptv");
	char *caid   = strdup("A0");
	char *source = strdup("P");
	char *param1 = strdup("B6");
	char *param3 = strdup("");
	char *param2 = strdup("");
	char *transponder = strdup(";w_pvrscan");

	info("w_pvrscan version %d\n\n", version);

	memset(&ext_ctrls, 0, sizeof(ext_ctrls));
	memset(&ext_ctrl, 0, sizeof(ext_ctrl));
	memset(&vcap, 0, sizeof(vcap));                    
	memset(&vfmt, 0, sizeof(vfmt));
	memset(&vin, 0, sizeof(vin));
	memset(&vout, 0, sizeof(vout));
	memset(&vf, 0, sizeof(vf));
	memset(&vtuner, 0, sizeof(vtuner));

	while (1) {
	  option_index = 0;
	  idx = 0;
	  for (i = 0; long_options[i].name; i++) {
	    if (!isalpha(long_options[i].val)) continue;
	    short_options[idx++] = long_options[i].val;
	    if (long_options[i].has_arg == required_argument)
	      short_options[idx++] = ':';
	    }
	  short_options[idx] = 0;
	  ch = getopt_long(argc, argv, short_options, long_options, &option_index);
	  if (ch == -1) break;

	  switch (ch) {
                case 0: break;

		case OptHelp:
			usage();
			return 0;

		case OptVdrver:
			vdrver = strtol(optarg, 0L, 0);
			break;

		case OptDebug:
			debug++;
			break;

		case OptRadio:
			radio=1;
			break;

		case OptRadioDev:
			radio=1;
			radio_dev = strtol(optarg, 0L, 0);
			break;

		case OptPlugin:
			plugin = strdup(optarg);
			break;

		case OptOutfile:
			outfile = strdup(optarg);
                        ofile++;
			break;

		case OptAdapter:
			dev = strtol(optarg, 0L, 0);
			if ((dev < 0) || (dev >= MAX_CARDS)) {
				usage();
				return -1;
				}
			break;

		case OptVerbose:
			verbosity++;
			break;

		case OptQuiet:
			if (--verbosity < 0) verbosity = 0;
			break;

		case '?':
			fprintf(stderr, "Unknown argument `%s'\n", argv[optind]);
			usage();
			return 1;
		}
	}
	if (optind < argc) {
		error("unknown arguments: ");
		while (optind < argc)
			error("%s ", argv[optind++]);
		error("\n");
		usage();
		return 1;
		}
        if (!memcmp(plugin, "analogtv", 8)) radio = 0; /* not supported by analogtv plugin - ignored */

/* BEGIN SEARCHING FIRST IVTV CARD */
      info("-_-_-_-_-_-_-Initializing TV card-_-_-_-_-_-_-_-_ \n");

      if (dev < 0) {
	info("Info: using IVTV adapter auto detection.\n");
        for (dev = 0; dev <= MAX_CARDS; dev++) {
          sprintf(device,"/dev/video%d",dev);
	  if ((fd = open(device, O_RDWR)) < 0)
            continue;
          if (ioctl(fd, VIDIOC_QUERYCAP, &vcap)!= 0) {
            close(fd);
            fd=-1;
	    }
          else {
            if (!memcmp(vcap.driver, "ivtv", 4)) {
              info("Info: Using ivtv device %s\n",device);
	      info("Info: Card is %s.\n", vcap.card);
	      if ((vcap.capabilities & V4L2_CAP_RADIO) && (memcmp(plugin, "analogtv", 8))) {
		/* FM radio and not analogtv plugin */
		radio = 1;
		info("Info: Activating FM radio channel search\n");
		}
	      break;
              }
            }
          }
	}
	else {
          sprintf(device,"/dev/video%d",dev);
	  if ((fd = open(device, O_RDWR)) < 0) {
          	fatal("CANNOT OPEN DEVICE. => EXITING.\n");
		return -1;
		} 
          if (ioctl(fd, VIDIOC_QUERYCAP, &vcap)!= 0) {
          	errorn("CANNOT QUERY DEVICE CAPABILITIES. => EXITING.\n");
		close(fd);
		return -1;
		} 
          if (!memcmp(vcap.driver, "ivtv", 4)) {
          	info("Info: Using ivtv device %s\n",device); 
		info("Info: Card is %s.\n", vcap.card);
		if ((vcap.capabilities & V4L2_CAP_RADIO) && (memcmp(plugin, "analogtv", 8))) {
		  /* FM radio and not analogtv */
		  radio = 1;
		  info("Info: Activating FM radio channel search\n");
		  }
		}
	  else {
		close(fd);
		fatal("DEVICE %s IS NO IVTV DEVICE. => EXITING.\n",device);
		return -1;		
		}
	  }		
/* END SEARCHING FIRST IVTV CARD */
	if (fd < 0) {
          fatal("NO VALID IVTV DEVICE FOUND. => EXITING.\n"); 
	  exit(1);
	  }
/* BEGIN CARD INITIALIZATION */
	if (debug) info("Info: Setting MPEG codec\n");

	/* setting new v4l2 ext mpeg ctrls */
	ext_ctrls.ctrl_class=V4L2_CTRL_CLASS_MPEG;

	if (debug) info("Info: Setting aspect\n");
	ext_ctrl.id=V4L2_CID_MPEG_VIDEO_ASPECT;
	ext_ctrl.value=V4L2_MPEG_VIDEO_ASPECT_16x9;
	ext_ctrls.controls=&ext_ctrl;
	ext_ctrls.count=1;
	doioctl(fd, VIDIOC_S_EXT_CTRLS, &ext_ctrls, "V4L2_CID_MPEG_VIDEO_ASPECT");

	if (debug) info("Info: Setting video bitrate mode\n");
	ext_ctrl.id=V4L2_CID_MPEG_VIDEO_BITRATE_MODE;
	ext_ctrl.value=V4L2_MPEG_VIDEO_BITRATE_MODE_CBR;
	ext_ctrls.controls=&ext_ctrl;
	ext_ctrls.count=1;
	doioctl(fd, VIDIOC_S_EXT_CTRLS, &ext_ctrls, "V4L2_CID_MPEG_VIDEO_BITRATE_MODE");

	if (debug) info("Info: Setting video bitrate\n");	
	ext_ctrl.id=V4L2_CID_MPEG_VIDEO_BITRATE;
	ext_ctrl.value=5000000;
	ext_ctrls.controls=&ext_ctrl;
	ext_ctrls.count=1;
	doioctl(fd, VIDIOC_S_EXT_CTRLS, &ext_ctrls, "V4L2_CID_MPEG_VIDEO_BITRATE");

	if (debug) info("Info: Setting video bitrate peak\n");		
	ext_ctrl.id=V4L2_CID_MPEG_VIDEO_BITRATE_PEAK;
	ext_ctrl.value=6000000;
	ext_ctrls.controls=&ext_ctrl;
	ext_ctrls.count=1;
	doioctl(fd, VIDIOC_S_EXT_CTRLS, &ext_ctrls, "V4L2_CID_MPEG_VIDEO_BITRATE_PEAK");

	if (debug) info("Info: Setting mpeg stream type\n");	
	ext_ctrl.id=V4L2_CID_MPEG_STREAM_TYPE;
	ext_ctrl.value=V4L2_MPEG_STREAM_TYPE_MPEG2_PS;
	ext_ctrls.controls=&ext_ctrl;
	ext_ctrls.count=1;
	doioctl(fd, VIDIOC_S_EXT_CTRLS, &ext_ctrls, "V4L2_CID_MPEG_STREAM_TYPE");

	if (debug) info("Info: Setting audio bitrate\n");	
	ext_ctrl.id=V4L2_CID_MPEG_AUDIO_L2_BITRATE;
	ext_ctrl.value=V4L2_MPEG_AUDIO_L2_BITRATE_320K;
	ext_ctrls.controls=&ext_ctrl;
	ext_ctrls.count=1;
	doioctl(fd, VIDIOC_S_EXT_CTRLS, &ext_ctrls, "V4L2_CID_MPEG_AUDIO_L2_BITRATE");

	if (debug) info("Info: Setting spatial filter\n");	
	ext_ctrl.id=V4L2_CID_MPEG_CX2341X_VIDEO_SPATIAL_FILTER_MODE;
	ext_ctrl.value=V4L2_MPEG_CX2341X_VIDEO_SPATIAL_FILTER_MODE_MANUAL;
	ext_ctrls.controls=&ext_ctrl;
	ext_ctrls.count=1;
	doioctl(fd, VIDIOC_S_EXT_CTRLS, &ext_ctrls, "V4L2_CID_MPEG_CX2341X_VIDEO_SPATIAL_FILTER_MODE");

	if (debug) info("Info: Setting temporal filter\n");	
	ext_ctrl.id=V4L2_CID_MPEG_CX2341X_VIDEO_TEMPORAL_FILTER_MODE;
	ext_ctrl.value=V4L2_MPEG_CX2341X_VIDEO_TEMPORAL_FILTER_MODE_MANUAL;
	ext_ctrls.controls=&ext_ctrl;
	ext_ctrls.count=1;
	doioctl(fd, VIDIOC_S_EXT_CTRLS, &ext_ctrls, "V4L2_CID_MPEG_CX2341X_VIDEO_TEMPORAL_FILTER_MODE");

	if (debug) info("Info: Setting median filter\n");	
	ext_ctrl.id=V4L2_CID_MPEG_CX2341X_VIDEO_MEDIAN_FILTER_TYPE;
	ext_ctrl.value=V4L2_MPEG_CX2341X_VIDEO_MEDIAN_FILTER_TYPE_OFF;
	ext_ctrls.controls=&ext_ctrl;
	ext_ctrls.count=1;
	doioctl(fd, VIDIOC_S_EXT_CTRLS, &ext_ctrls, "V4L2_CID_MPEG_CX2341X_VIDEO_MEDIAN_FILTER_TYPE");

	if (debug) info("Info: Setting video pulldown\n");	
	ext_ctrl.id=V4L2_CID_MPEG_VIDEO_PULLDOWN;
	ext_ctrl.value=0;
	ext_ctrls.controls=&ext_ctrl;
	ext_ctrls.count=1;
	doioctl(fd, VIDIOC_S_EXT_CTRLS, &ext_ctrls, "V4L2_CID_MPEG_VIDEO_PULLDOWN");
	if (debug) info("Info: End setting codec.\n");	

	/* end setting mpeg codec */
	
	vfmt.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
	if (debug) info("Info: Setting format\n");
	doioctl(fd, VIDIOC_G_FMT, &vfmt, "VIDIOC_G_FMT");
	vfmt.fmt.pix.width = 720;
	vfmt.fmt.pix.height = 576;
	doioctl(fd, VIDIOC_S_FMT, &vfmt,"VIDIOC_S_FMT");

	/* now setting only ctrl_class_user ioctls */
	ext_ctrls.ctrl_class=V4L2_CTRL_CLASS_USER;

	if (debug) info("Info: Setting brightness\n");	
	ext_ctrl.id=V4L2_CID_BRIGHTNESS;
	ext_ctrl.value=135;
	ext_ctrls.controls=&ext_ctrl;
	ext_ctrls.count=1;
	doioctl(fd, VIDIOC_S_EXT_CTRLS, &ext_ctrls, "V4L2_CID_BRIGHTNESS");

	if (debug) info("Info: Setting contrast\n");
	ext_ctrl.id=V4L2_CID_CONTRAST;
	ext_ctrl.value=64;
	ext_ctrls.controls=&ext_ctrl;
	ext_ctrls.count=1;
	doioctl(fd, VIDIOC_S_EXT_CTRLS, &ext_ctrls, "V4L2_CID_CONTRAST");

	if (debug) info("Info: Setting saturation\n");
	ext_ctrl.id=V4L2_CID_SATURATION;
	ext_ctrl.value=64;
	ext_ctrls.controls=&ext_ctrl;
	ext_ctrls.count=1;
	doioctl(fd, VIDIOC_S_EXT_CTRLS, &ext_ctrls, "V4L2_CID_SATURATION");

	if (debug) info("Info: Setting hue\n");
	ext_ctrl.id=V4L2_CID_HUE;
	ext_ctrl.value=0;
	ext_ctrls.controls=&ext_ctrl;
	ext_ctrls.count=1;
	doioctl(fd, VIDIOC_S_EXT_CTRLS, &ext_ctrls, "V4L2_CID_HUE");

	if (debug) info("Info: Setting audio volume\n");
	ext_ctrl.id = V4L2_CID_AUDIO_VOLUME;
	ext_ctrl.value = 59636;
	ext_ctrls.controls=&ext_ctrl;
	ext_ctrls.count=1;
	doioctl(fd, VIDIOC_S_EXT_CTRLS, &ext_ctrls, "V4L2_CID_AUDIO_VOLUME");

	if (debug) info("Info: Setting audio mute\n");
	ext_ctrl.id = V4L2_CID_AUDIO_MUTE;
	ext_ctrl.value = 0;
	ext_ctrls.controls=&ext_ctrl;
	ext_ctrls.count=1;
	doioctl(fd, VIDIOC_S_EXT_CTRLS, &ext_ctrls, "V4L2_CID_AUDIO_MUTE");

	if (debug) info("Info: Setting video standard\n");
	std = V4L2_STD_PAL;
        doioctl (fd, VIDIOC_S_STD, &std,"VIDIOC_S_STD");
	vin.index = 0;
	if (debug) info("Info: Searching TV tuner\n");
	while (ioctl(fd, VIDIOC_ENUMINPUT, &vin) >= 0) {
	  if (!memcmp(vin.name, "Tuner", 5)) {
            input=vin.index;
            if (debug) info("Info: TV Tuner is input %d\n",input);
            break;             
            }
	  vin.index++;
	  }
	if (debug) info("Info: Setting video input to TV tuner\n");
	doioctl(fd, VIDIOC_S_INPUT, &input, "VIDIOC_S_INPUT");
	vout.index = 0; 
	if (debug) info("Info: Getting other video inputs\n");       
	while (ioctl(fd, VIDIOC_ENUMOUTPUT, &vout) >= 0) {
	  if (!memcmp(vout.name, "S-Video", 7)) {
            output=vout.index;
            break;             
            }
	  if (!memcmp(vout.name, "Composite", 9)) {
            output=vout.index;
            break;             
            }		  
	  vout.index++;
	  }
	info("Info: Initialization done.\n");
        info("-_-_-_-_-_-_-_-_-_-_-_-_-_-_-_-_-_-_-_-_-_-_-_ \n");
/* END CARD INITIALIZATION */

/* Card is now perfectly configured for viewing PAL TV,
   let's scan all frequencies now.. */

  info("\nScanning.. please wait. This will take up to five minutes!\n\n");
  if (ofile > 0) {
     conf = fopen(outfile,"w");
    }
  if (!memcmp(plugin, "analogtv", 8)) {
    caid   = strdup("A0");
    source = strdup("C");
    param1 = strdup("B6");
    switch(dev) {
	case 0: param2 = strdup("C0"); break;
	case 1: param2 = strdup("C12"); break;
	case 2: param2 = strdup("C23"); break;
	case 3: param2 = strdup("C34"); break;
	case 4: param2 = strdup("C45"); break;
	case 5: param2 = strdup("C56"); break;
	case 6: param2 = strdup("C67"); break;
	case 7: param2 = strdup("C78"); break;
	case 8: param2 = strdup("C89"); break;
	}
    switch(input) {
	case 0: param3 = strdup("D0"); break;
	case 1: param3 = strdup("D12"); break;
	case 2: param3 = strdup("D23"); break;
	case 3: param3 = strdup("D34"); break;
	case 4: param3 = strdup("D45"); break;
	case 5: param3 = strdup("D56"); break;
	case 6: param3 = strdup("D67"); break;
	case 7: param3 = strdup("D78"); break;
	case 8: param3 = strdup("D89"); break;
       }
    }
  if (vdrver < 3) {
    caid = strdup ("32001");
    transponder = strdup("");
    }
  if (!memcmp(plugin, "pvrinput", 8)) {
    caid   = strdup("A1");
    source = strdup("C");
    param1 = strdup("I999");
    switch(input) {
	case 0: param3 = strdup("C0"); break;
	case 1: param3 = strdup("C12"); break;
	case 2: param3 = strdup("C23"); break;
	case 3: param3 = strdup("C34"); break;
	case 4: param3 = strdup("C45"); break;
	case 5: param3 = strdup("C56"); break;
	case 6: param3 = strdup("C67"); break;
	case 7: param3 = strdup("C78"); break;
	case 8: param3 = strdup("C89"); break;
	   }
    }

/* now scanning europe-west, CCIR_I_III, K02..K05, S01..S03 */
vf.type=V4L2_TUNER_ANALOG_TV;
for (i=0;i<=5;i++) {			
  vf.tuner = 0;
  vf.frequency=(16*(48250 + i*7000))/1000;          
  doioctl(fd, VIDIOC_S_FREQUENCY, &vf,"VIDIOC_S_FREQUENCY");
  if (i<=2)
    sprintf(sendername, "k0%d", i+2);
  else
    sprintf(sendername, "S0%d", i-2);
  info("%3.2fMHz", (double) vf.frequency/16);
  tmr=0;
  usleep(TV_TUNING_TIME);
  while (getSignalstatus(fd) < 1)
    {
    usleep(RETRY_TIME);
    tmr++;
    if (tmr > MAX_TRY-1) {
      break;
      }
    }
  if (tmr > MAX_TRY-1) {
    info("\n");
    continue;
    }
  info(": found new channel %s\n", sendername);    
  count++;
  fprintf(conf,"%s%s:%d:%s%s%s:%s:0:301:300:305:%s:%d:0:0:0\n", sendername,\
      transponder, vf.frequency*1000/16, param1, param2, param3, source, caid, vf.frequency);
  if (ofile)
    fflush(conf);      
  }

/* now scanning europe-west, CCIR_SL_SH, SE01..SE20 && CCIR_I_III, k05..k12 */

for (i=0;i<=27;i++) {			
  vf.tuner = 0;
  vf.frequency=(16*(105250 + i*7000))/1000;          
  doioctl(fd, VIDIOC_S_FREQUENCY, &vf,"VIDIOC_S_FREQUENCY");
  if (i<=9)
    sprintf(sendername, "SE%d", i+1);
  if (i>9)
    sprintf(sendername, "k0%d", i-5);
  if (i>14)
    sprintf(sendername, "k%d", i-5);
  if (i>17)
    sprintf(sendername, "SE%d", i-7);
  tmr=0;
  info("%3.2fMHz", (double) vf.frequency/16);
  usleep(TV_TUNING_TIME);
  while (getSignalstatus(fd) < 1)
    {
    usleep(RETRY_TIME);
    tmr++;
    if (tmr > MAX_TRY-1) {
      break;
      }
    }
  if (tmr > MAX_TRY-1) {
    info("\n");
    continue;
    }
  info(": found new channel %s\n", sendername);    
  count++;
  fprintf(conf,"%s%s:%d:%s%s%s:%s:0:301:300:305:%s:%d:0:0:0\n", sendername,\
      transponder, vf.frequency*1000/16, param1, param2, param3, source, caid, vf.frequency);
  if (ofile)
    fflush(conf);  
  }

/* now scanning europe-west, CCIR_H, S21..S41 && UHF k21..k69 */

  for (i=0;i<=69;i++) {
  vf.tuner = 0;
  vf.frequency=(16*(303250 + i*8000))/1000;           
  doioctl(fd, VIDIOC_S_FREQUENCY, &vf,"VIDIOC_S_FREQUENCY");
  if (i<=20)
    sprintf(sendername, "S%d", i+21);
  else
    sprintf(sendername, "k%d", i);
  tmr=0;
  info("%3.2fMHz", (double) vf.frequency/16);
  usleep(TV_TUNING_TIME);
  while (getSignalstatus(fd) < 1)
    {
    usleep(RETRY_TIME);
    tmr++;
    if (tmr > MAX_TRY-1) {
      break;
      }
    }
  if (tmr > MAX_TRY-1) {
    info("\n");
    continue;
    }
  info(": found new channel %s\n", sendername);
  count++;
  fprintf(conf,"%s%s:%d:%s%s%s:%s:0:301:300:305:%s:%d:0:0:0\n", sendername,\
      transponder, vf.frequency*1000/16, param1, param2, param3, source, caid, vf.frequency);
  if (ofile)
    fflush(conf);
  }
  vin.index = 0;
  while (ioctl(fd, VIDIOC_ENUMINPUT, &vin) >= 0) {
    if (!memcmp(vin.name, "Tuner", 5)) {
      vin.index++;
      continue;             
      }
    switch(vin.index) {
	case 0: param3 = (!memcmp(plugin, "pvrinput", 8))? strdup("C0"): strdup("D0");  break;
	case 1: param3 = (!memcmp(plugin, "pvrinput", 8))? strdup("C12"):strdup("D12"); break;
	case 2: param3 = (!memcmp(plugin, "pvrinput", 8))? strdup("C23"):strdup("D23"); break;
	case 3: param3 = (!memcmp(plugin, "pvrinput", 8))? strdup("C34"):strdup("D34"); break;
	case 4: param3 = (!memcmp(plugin, "pvrinput", 8))? strdup("C45"):strdup("D45"); break;
	case 5: param3 = (!memcmp(plugin, "pvrinput", 8))? strdup("C56"):strdup("D56"); break;
	case 6: param3 = (!memcmp(plugin, "pvrinput", 8))? strdup("C67"):strdup("D67"); break;
	case 7: param3 = (!memcmp(plugin, "pvrinput", 8))? strdup("C78"):strdup("D78"); break;
	case 8: param3 = (!memcmp(plugin, "pvrinput", 8))? strdup("C89"):strdup("D89"); break;
       }
    if (!memcmp(plugin, "ptv", 3)) {
      if (!memcmp(vin.name, "Composite 0", 11)) param3 = strdup("D0");
      if (!memcmp(vin.name, "Composite 1", 11)) param3 = strdup("D12");
      if (!memcmp(vin.name, "Composite 2", 11)) param3 = strdup("D23");
      if (!memcmp(vin.name, "Composite 3", 11)) param3 = strdup("D34");
      if (!memcmp(vin.name, "Composite 4", 11)) param3 = strdup("D56");
      if (!memcmp(vin.name, "S-Video 3", 12)) continue;
      if (!memcmp(vin.name, "S-Video 0", 12)) param3 = strdup("D67");
      if (!memcmp(vin.name, "S-Video 1", 12)) param3 = strdup("D78");
      if (!memcmp(vin.name, "S-Video 2", 12)) param3 = strdup("D89");
      }
    fprintf(conf,"%s%s:%d:%s%s%s:%s:0:301:300:305:%s:%d:0:0:0\n", vin.name,\
      transponder, 1, param1, param2, param3, source, caid, 9000+vin.index);
    if (ofile)
	fflush(conf);
    vin.index++;   
    }    
  if ((ofile) && (!radio)) {
    fflush(conf);
    fclose(conf);
    ofile = 0;
    }
  close(fd);
  info("Found total %d TV Stations.\n\n", count);


  if (radio) {
	/* BEGIN SEARCHING FIRST IVTV RADIO CARD */
      info("-_-_-_-_-_-_-Initializing radio card-_-_-_-_-_-_-_-_ \n");
      count = 0;
      if (radio_dev < 0) {
	info("Info: using IVTV adapter auto detection.\n");
        for (radio_dev = 0; radio_dev <= MAX_CARDS; radio_dev++) {
          sprintf(device,"/dev/radio%d",radio_dev);
	  if ((fd = open(device, O_RDWR)) < 0)
            continue;
          if (ioctl(fd, VIDIOC_QUERYCAP, &vcap)!= 0) {
            close(fd);
            fd=-1;
	    }
          else {
            if ((!memcmp(vcap.driver, "ivtv", 4)) && (vcap.capabilities & V4L2_CAP_RADIO)) {
              info("Info: Using ivtv radio device %s\n",device);
	      info("Info: Card is %s.\n", vcap.card);
	      break;
              }
            }
          }
	}
	else {
          sprintf(device,"/dev/radio%d",radio_dev);
	  if ((fd = open(device, O_RDWR)) < 0) {
          	fatal("CANNOT OPEN DEVICE. => EXITING.\n");
		return -1;
		} 
          if (ioctl(fd, VIDIOC_QUERYCAP, &vcap)!= 0) {
          	errorn("CANNOT QUERY DEVICE CAPABILITIES. => EXITING.\n");
		close(fd);
		return -1;
		} 
          if (!memcmp(vcap.driver, "ivtv", 4) && (vcap.capabilities & V4L2_CAP_RADIO)) {
          	info("Info: Using ivtv device %s\n",device); 
		info("Info: Card is %s.\n", vcap.card);
		}
	  else {
		info("DEVICE %s IS NO IVTV RADIO DEVICE. => EXITING.\n",device);
		close(fd);
		return -1;		
		}
	  }
	  if (vcap.capabilities & V4L2_TUNER_CAP_STEREO) {
		stereo++;
		info("Info: This device is stereo capable.\n");
		}
	  vin.index = 0;
	  if (debug) info("Info: Searching tuner input\n");
	  while (ioctl(fd, VIDIOC_ENUMINPUT, &vin) >= 0) {
	    if (!memcmp(vin.name, "Tuner", 5)) {
              input=vin.index;
              if (debug) info("Info: Tuner is input %d\n",input);
              break;             
              }
	    vin.index++;
	    }
	  if (!memcmp(plugin, "pvrinput", 8)) {
             caid   = strdup("A1");
             source = strdup("C");
             param1 = strdup("I999");
             switch(input) {
			case 0: param3 = strdup("C0");  break;
			case 1: param3 = strdup("C12"); break;
			case 2: param3 = strdup("C23"); break;
			case 3: param3 = strdup("C34"); break;
			case 4: param3 = strdup("C45"); break;
			case 5: param3 = strdup("C56"); break;
			case 6: param3 = strdup("C67"); break;
			case 7: param3 = strdup("C78"); break;
			case 8: param3 = strdup("C89"); break;
	   		}
		}
	  else
		param3 = strdup("D45"); 		/* ptv plugin tuner 0 */
	doioctl(fd, VIDIOC_G_TUNER, &vtuner, "VIDIOC_G_TUNER");		
	if (stereo) vtuner.audmode=V4L2_TUNER_MODE_STEREO;
	doioctl(fd, VIDIOC_S_TUNER, &vtuner, "VIDIOC_S_TUNER");			
	info("Info: Initialization done.\n");
        info("-_-_-_-_-_-_-_-_-_-_-_-_-_-_-_-_-_-_-_-_-_-_-_ \n\n");
		
	/* END SEARCHING FIRST IVTV RADIO CARD */
#ifndef TVTEST
	if (ofile > 0) {
	    conf = fopen(outfile,"w");
	    }
#endif
	for (i = 1400; i <= 1728; i++) { /* (87.5MHz .. 108MHz) x 16 steps/MHz */
		vf.tuner = 0; vtuner.signal = 0;
                vf.frequency = i;
		doioctl(fd, VIDIOC_S_FREQUENCY, &vf,"VIDIOC_S_FREQUENCY");
		usleep(RADIO_TUNING_TIME);
		doioctl(fd, VIDIOC_G_TUNER, &vtuner, "VIDIOC_G_TUNER");
		strength = vtuner.signal;
		if (strength >= MIN_SIGNAL_STRENGTH) {
			info("%3.2fMHz, signal strength %#3.1f%%\n",\
					(double) vf.frequency/16, \
					(double) strength/655.36);
			fprintf(conf,"FM %#3.2fMHz%s:%d:%s%s%s:%s:0:0:300:305:%s:%d:0:0:0\n", \
      					(double) formatFreq((uint) (62.5 * vf.frequency))/1000, \
					transponder, \
					(uint) (62.5 * vf.frequency),\
					param1, param2, param3,\
					source, caid, vf.frequency);
			if (ofile)
			    fflush(conf);
			count++;
			}
		else {
			info("%3.2fMHz\n", (double) vf.frequency/16);
			}

		} /* end of FM radio scan loop */
	if (ofile) {
		fflush(conf);
		fclose(conf);
		}
        close(fd);
	info("Found total %d FM Radio Stations.\n\n", count);
  }

  info("REMEMBER: Please use *always* the latest ivtv versions, plugin versions and VDR versions,\n");
  info("otherwise you may have wrong syntax / non working software!!\n\n");
  exit(0);
}
